// NeoPixel Ring simple sketch (c) 2013 Shae Erisson released under
// the GPLv3 license to match the rest of the AdaFruit NeoPixel
// library

#include <Adafruit_NeoPixel.h>
#ifdef __AVR__
  #include <avr/power.h>
#endif

#include "react.h"

// Which pin on the Arduino is connected to the NeoPixels?
// On a Trinket or Gemma we suggest changing this to 1
#define PIN            12

#define BUTTON_PIN     10

// How many NeoPixels are attached to the Arduino?
#define NUMPIXELS      (24+7)

//#define BLUE_WAVE 1
#define BLUE_FAST 1

// When we setup the NeoPixel library, we tell it how many pixels, and
// which pin to use to send signals. Note that for older NeoPixel
// strips you might need to change the third parameter--see the
// strandtest example for more information on possible values.
Adafruit_NeoPixel pixels = Adafruit_NeoPixel(
    NUMPIXELS, PIN, NEO_GRBW + NEO_KHZ800);

void setup()
{
    // This is for Trinket 5V 16MHz, you can remove these three lines if
    // you are not using a Trinket
    #if defined (__AVR_ATtiny85__)
    if (F_CPU == 16000000) clock_prescale_set(clock_div_1);
    #endif
    // End of trinket special code
    
    pixels.begin(); // This initializes the NeoPixel library.

    pinMode(BUTTON_PIN, INPUT_PULLUP);
}

float wave(float theta)
{
    return (cos(theta * M_PI / 180.) + 1.) / 2.;
}

// convert from 0..1 to 0..255
inline int colorComponent(float c)
{
    int i = (int)round(c * 255);
    if (i < 0) {
        return 0;
    } else if (i > 255) {
        return 255;
    } else {
        return i;
    }
}

// set the color of LED i. The color components are in the range 0..1.
void setPixel(int i, Color & c)
{
    int r = colorComponent(c.r);
    int g = colorComponent(c.g);
    int b = colorComponent(c.b);
    int w = min(min(r,g), b);
    r -= w;
    g -= w;
    b -= w;
    pixels.setPixelColor(i, pixels.Color(r,g,b,w));
}

inline float smoothstep(float s, float lo, float hi)
{
    s = (s - lo) / (hi - lo);
    if (s < 0)
        return 0;
    else if (s > 1)
        return 1;
    else
        return 3*s*s - 2*s*s*s;
}

// a red "turning on" effect
// x and on go from 0..1
// 
// 0..1 black
// 1..2 black -> red
// 2..3 red -> white
// 3..4 white

void modulate(float x, float on, Color & c)
{
    float z = on*3 + x;
    float r1 = smoothstep(z, 1, 2);
    float r2 = smoothstep(z, 2, 3);
    float red = r1 * (1 - r2);
    c.r = c.r * r1 + red * 0.5;
    c.g *= r2;
    c.b *= r2;
}

void loop()
{
    
    // For a set of NeoPixels the first NeoPixel is 0, second is 1, all
    // the way up to the count of pixels minus one.

    int time = 0;
    float dc; // color component spread
    float dl; // LED spread

    #if BLUE_FAST
    #define NUM_STREAKS 4 // maximum
    Streak streaks[NUM_STREAKS];
    int s;
    for (s = 0; s < NUM_STREAKS; ++s)
        streaks[s].clear();
    float buffer[24]; // brightness of streaks
    #endif
    
    if (true) { // rainbow
        dc = 1.;
        dl = 0.5;
    } else if (true) { // white
        dc = 0.;
        dl = 0.;
    }
    while (true) {
        int button = digitalRead(BUTTON_PIN);
        const float d_on = 0.01;
        static float on = 0;
        if (button) {
            on += d_on;
            if (on > 1) on = 1;
        } else {
            on -= d_on;
            if (on < 0) on = 0;
        }
        float ss_on = smoothstep(on, 0, 1);

        #if BLUE_WAVE
        for (int i = 0; i < NUMPIXELS; i++) {
            float x = 1 - (i + 0.5) / NUMPIXELS;
            Color c;
            
            float theta = time * 5;

            // old blue and white continuous wave
            int j = i;
            if (j > 1 and j <= 6)
                j = 1;
            c.r = wave((theta + j*dl) * 120);
            c.r = pow(c.r, 16);
            c.g = c.r;
            c.b = 1;
            
            modulate(x, ss_on, c);
            setPixel(i, c);
        }
        #endif
        
        //------------------------------------------------------
        #if BLUE_FAST
        // blue with racing white bits

        const float tsoft = 10; // fade in / fade out time
        const float xlen = 3; // how many LEDs long
        const float xsoft = 3; // fade in / fade out distance (LEDs)
        const int speed = 3; // LEDs per time step
        
        int i;
        for (i = 0; i <= 6; ++i) {
            int phi = 0; // center
            if (i > 0)
                phi = -30; // inner ring
            float theta = time * 1.5;
            
            Color c;
            c.r = 0.5 * pow(wave(theta + phi), 64);
            c.g = c.r;
            c.b = 1;
                
            float x = 1 - (i + 0.5) / NUMPIXELS;
            modulate(x, ss_on, c);
            setPixel(i, c);
        }
        
        // outer ring
        // first, render the current picture of our streaks without rotation
        for (i = 0; i < 24; ++i) {
            buffer[i] = 0;
        }
        for (s = 0; s < NUM_STREAKS; ++s) {
            Streak & streak = streaks[s];
            if (not streak.active())
                continue;
            float s_alpha = 0.2 *
                smoothstep(time, streak.start - tsoft, streak.start + tsoft) *
                smoothstep(time, streak.finish + tsoft, streak.finish - tsoft);
            int x0 = streak.angle;
            float x1 = x0 + xlen;
            for (i = x0 - xsoft - 1; i < x1 + xsoft + 1; ++i) {
                float alpha = s_alpha *
                    smoothstep(i, x0 - xsoft, x0 + xsoft) *
                    smoothstep(i, x1 + xsoft, x1 - xsoft);
                float & white = buffer[i % 24];
                white = white * (1 - alpha) + alpha;
            }
        }
        
        for (i = 7; i < NUMPIXELS; ++i) {
            // then rotate that picture to render the streaks.
            
            float white = buffer[(i + time*2) % 24];
            //const float theta = (i - 7.f) / 24.f;
            Color c;
            c.r = c.g = white;
            c.b = 1;

            float x = 1 - (i + 0.5) / NUMPIXELS;
            modulate(x, ss_on, c);
            setPixel(i, c);
        }
        
        // Manage animation. First, look for streaks that are done
        // and delete them.
        for (s = 0; s < NUM_STREAKS; ++s) {
            Streak & streak = streaks[s];
            if (!streak.active())
                continue;
            if (time > streak.finish + tsoft)
                streak.clear();
        }
        // Then, sometimes randomly start a new one.
        if (random(40) == 0) {
            s = random(NUM_STREAKS);
            Streak & streak = streaks[s];
            if (not streak.active()) {
                streak.angle = random(24);
                streak.start = time + tsoft;
                streak.finish = streak.start + random(12, 48) / speed;
            }
        }
        
        #endif
        
        // This sends the updated pixel color to the hardware.
        pixels.show();
        
        time += 1;
        
        const int delayval = 5; // msec

        // Delay for a period of time (in milliseconds).
        delay(delayval);
    }
}
